#include "SteamTracking.h"



std::string SteamTracking::getDeviceTypeAsString(vr::ETrackedDeviceClass device_class)
{
	std::string type;
	switch (device_class)
	{
	case vr::TrackedDeviceClass_Controller:        type = "Controller"; break;
	case vr::TrackedDeviceClass_HMD:               type = "HMD"; break;
	case vr::TrackedDeviceClass_Invalid:           type = "Invalid"; break;
	case vr::TrackedDeviceClass_GenericTracker:    type = "GenericTracker"; break;
	case vr::TrackedDeviceClass_TrackingReference: type = "TrackingReference"; break;
	default:                                       type = "Unknown"; break;
	}
	return type;
}

glm::mat3x4* SteamTracking::convertMatrix(vr::HmdMatrix34_t in)
{
	return new glm::mat3x4(in.m[0][0], in.m[0][1], in.m[0][2], in.m[0][3],
		in.m[1][0], in.m[1][1], in.m[1][2], in.m[1][3],
		in.m[2][0], in.m[2][1], in.m[2][2], in.m[2][3]	);
}



SteamTracking::SteamTracking()
{
	// Loading the SteamVR Runtime
	vr::EVRInitError eError = vr::VRInitError_None;
	vr_system_ = vr::VR_Init(&eError, vr::VRApplication_Overlay);

	if (eError != vr::VRInitError_None)
	{
		vr_system_ = nullptr;
		std::cerr <<"Unable to init VR runtime: "<< vr::VR_GetVRInitErrorAsEnglishDescription(eError) << std::endl;
	}

}


SteamTracking::~SteamTracking()
{
	if (vr_system_)
	{
		vr::VR_Shutdown();
		vr_system_ = nullptr;
	}
}




void SteamTracking::printVRDevices() {
	std::cout << "List of available VR devices:" << std::endl;
	if (!vr_system_) {
		std::cout << "\tNot connected to VR System, no devices";

	}
	for (int nDevice = 0; nDevice < vr::k_unMaxTrackedDeviceCount; ++nDevice)
	{
		vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);
		if (device_class == vr::TrackedDeviceClass_Invalid)
		{
			continue;
		}

		const uint32_t size = 1024;
		char serial[size];
		vr_system_->GetStringTrackedDeviceProperty(nDevice, vr::ETrackedDeviceProperty::Prop_SerialNumber_String, serial, size);

		std::cout << "\tDevice " << nDevice << ":" << std::endl;
		std::cout << "\t\tType: " << getDeviceTypeAsString(device_class) << std::endl;
		std::cout << "\t\tSerial: " << serial << std::endl;
	}
}


void SteamTracking::printMatrix(const vr::HmdMatrix34_t &htm)
{
	for (int rows = 0; rows < 3; ++rows)
	{
		std::cout << (rows == 0 ? "[" : "") << "\t";

		for (int cols = 0; cols < 4; ++cols)
		{
			printf("%.3f\t", htm.m[rows][cols]);
			//std::cout << htm.m[rows][cols] << "\t";
		}
		std::cout << (rows == 2 ? "]" : "") << std::endl;
	}
}

glm::mat3x4* SteamTracking::getTransformationForDevice(int nDevice)
{
	if (!vr_system_) {
		return new glm::mat3x4();
	}

	vr_system_->GetDeviceToAbsoluteTrackingPose(vr::TrackingUniverseRawAndUncalibrated, 0, tracked_device_poses_, vr::k_unMaxTrackedDeviceCount);

	if (!tracked_device_poses_[nDevice].bPoseIsValid)
	{
		if (debug) std::cout << " Device " << nDevice << "Pois is Invalid";
		return new glm::mat3x4();
	}

	vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);

	if (device_class == vr::TrackedDeviceClass_Invalid)
	{
		if (debug) std::cout << " Device " << nDevice << "ist Invalid";
		return new glm::mat3x4();
	}

	vr::HmdMatrix34_t htm = tracked_device_poses_[nDevice].mDeviceToAbsoluteTracking;

	if (debug)
	{
		std::cout << "Pose of device " << nDevice << " (" << getDeviceTypeAsString(device_class)  << ")" << std::endl;
		printMatrix(htm);
	}

	return convertMatrix(htm);
}
